#include "dsdpdatamat_impl.h"
#include "dsdpdatamat.h"
#include "dsdpsys.h"
#include "dsdpbasictypes.h"
/*!
\file dsdpdatamat.c
\brief Call an implementation of the data matrix operations.
*/
#define DSDPNoOperationError(a);  { DSDPSETERR1(1,"Data natrix type: %s, Operation not defined\n",(a).dsdpops->matname); }
#define DSDPChkDataError(a,b);  { if (b){ DSDPSETERR1(b,"Data natrix type: %s,\n",(a).dsdpops->matname);} }


static struct  DSDPDataMat_Ops dsdpdatamatdefault;

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatSetData"
/*!
\fn int64 DSDPDataMatSetData(DSDPDataMat *A, struct DSDPDataMat_Ops* ops,  void*data);

\brief Set the opaque pointer and function pointers to the matrix.
\param A symmetric data matrix
\param ops pointer to a structure of function pointers
\param data pointer to a matrix structure

*/
int64 DSDPDataMatSetData(DSDPDataMat *A, struct DSDPDataMat_Ops* ops,  void*data){
  int64 info;
  DSDPFunctionBegin;
  (*A).dsdpops=ops;
  (*A).matdata=data;
  if (ops==NULL){
    (*A).dsdpops=&dsdpdatamatdefault;   
  }
  info = DSDPDataMatOpsInitialize(&dsdpdatamatdefault); DSDPCHKERR(info);
  info=DSDPDataMatTest(*A);DSDPCHKERR(info);
  DSDPFunctionReturn(0); 
}

static char datamatnoname[20]="MATRIX NOT SET";
#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatOpsInitialize"
/*!
\fn int64 DSDPDataMatOpsInitialize(struct  DSDPDataMat_Ops* dops);
\brief Initialize the table of function pointers for SDP Data matrices.

\param dops pointer to a structure of function pointers.
*/
int64 DSDPDataMatOpsInitialize(struct  DSDPDataMat_Ops* dops){
  DSDPFunctionBegin;
  if (dops==NULL) return 0;
  dops->matfactor1=0;
  dops->matfactor2=0;
  dops->matgetrank=0;
  dops->matgeteig=0;
  dops->matvecvec=0;
  dops->matdot=0;
  dops->mataddrowmultiple=0;
  dops->mataddallmultiple=0;
  dops->matdestroy=0;
  dops->matview=0;
  dops->matrownz=0;
  dops->matnnz=0;
  dops->matfnorm2=0;
  dops->id=0;
  dops->matname=datamatnoname;
  DSDPFunctionReturn(0); 
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatInitialize" 
/*!
\fn int64 DSDPDataMatInitialize(DSDPDataMat *A);

\brief Set pointers to NULL;
\param A symmetric data matrix

*/
int64 DSDPDataMatInitialize(DSDPDataMat *A){
  int64 info;
  DSDPFunctionBegin;
  info = DSDPDataMatOpsInitialize(&dsdpdatamatdefault); DSDPCHKERR(info);
  info = DSDPDataMatSetData(A, &dsdpdatamatdefault,0); DSDPCHKERR(info);
  DSDPFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatTest"
/*!
\fn int64 DSDPDataMatTest(DSDPDataMat A);

\brief Test validity of matrix
\param A symmetric data matrix

*/
int64 DSDPDataMatTest(DSDPDataMat A){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops==0 || A.dsdpops==&dsdpdatamatdefault){
  } else if (A.dsdpops->mattest){
    info=(A.dsdpops->mattest)(A.matdata); DSDPChkDataError(A,info);
  } else {
    /*
    DSDPNoOperationError(A);
    */
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatGetType"
int64 DSDPDataMatGetType(DSDPDataMat A, int64 *id){
  DSDPFunctionBegin;
  *id=A.dsdpops->id;
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatGetRank"
/*!
\fn int64 DSDPDataMatGetRank(DSDPDataMat A, int64 *rank, int64 n);

\brief Get the number of nonzero eigenvalues/eigenvectors for the matrix
\param A symmetric data matrix
\param rank number of nonzero eigenvalues and vectors.
\param n rows and columns in matrix.

*/
int64 DSDPDataMatGetRank(DSDPDataMat A, int64 *rank, int64 n){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops->matgetrank){
    info=(A.dsdpops->matgetrank)(A.matdata,rank,n); DSDPChkDataError(A,info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatCountNonzeros"
/*!
\fn int64 DSDPDataMatCountNonzeros(DSDPDataMat A, int64 *nnz, int64 n);

\brief Compute the square of the Frobenius norm 
\param A symmetric data matrix.
\param n dimension of the matrix.
\param *nnz nonzeros in matrix.

Used to identify which of a few strategies to compute Hessian.
*/
int64 DSDPDataMatCountNonzeros(DSDPDataMat A, int64 *nnz, int64 n){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops->matnnz){
    info=(A.dsdpops->matnnz)(A.matdata,nnz,n); DSDPChkDataError(A,info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatFNorm2"
/*!
\fn int64 DSDPDataMatFNorm2(DSDPDataMat A, int64 n, double *fnorm2);

\brief Compute the square of the Frobenius norm 
\param A symmetric data matrix.
\param n dimension of the matrix.
\param fnorm2 square of norm.

Used to scale the problem
*/
int64 DSDPDataMatFNorm2(DSDPDataMat A, int64 n, double *fnorm2){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops->matfnorm2){
    *fnorm2=0.0;
    info=(A.dsdpops->matfnorm2)(A.matdata,n,fnorm2); DSDPChkDataError(A,info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatGetEig"
/*!
\fn int64 DSDPDataMatGetEig(DSDPDataMat A, int64 rr, SDPConeVec V, DSDPIndex S, double *eigenvalue);

\brief Get an eigenvalue/vector pair.
\param A symmetric data matrix
\param rr identifies which pair of eigs,  (0 <= rr < rank)
\param V Eigenvector
\param S Indentifies sparsity pattern in V.
\param eigenvalue the scalar associated with the vector.
These pairs do not have to be eigenvalues and eigenvectors.  What matters
is that the matrix is a sum of the outer products of these vectors.
That is, A = sum (rr * V *V') 
\sa DSDPDataMatGetRank()
*/
int64 DSDPDataMatGetEig(DSDPDataMat A, int64 rr, SDPConeVec V, DSDPIndex S, double *eigenvalue){
  int64 info,n;
  double *vv;
  DSDPFunctionBegin;
  if (A.dsdpops->matgeteig){
    info=SDPConeVecGetArray(V,&vv); DSDPCHKERR(info);
    info=SDPConeVecGetSize(V,&n); DSDPCHKERR(info);
    info=(A.dsdpops->matgeteig)(A.matdata,rr, eigenvalue, vv,n,S.indx+1,S.indx); DSDPChkDataError(A,info);
    info=SDPConeVecRestoreArray(V,&vv); DSDPCHKERR(info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatFactor"
/*!
\fn int64 DSDPDataMatFactor(DSDPDataMat A, SDPConeVec W, double*dworknn, int64 nn0, double *dwork3n, int64 nd, int64* iwork, int64 ni){


\brief Do eigenvalue/vector or other factorization.
\param A symmetric data matrix
\param W work vector
\param dworknn work array
\param nn0 length of dworknn array.
\param dwork3n work array
\param nd length of work array
\param iwork work array
\param ni length of iwork array

This routine is called once during the DSDPSetUp routine.

\sa DSDPDataMatGetRank()
\sa DSDPDataMatGetEig()
*/
int64 DSDPDataMatFactor(DSDPDataMat A, SDPConeVec W, double*dworknn, int64 nn0, 
		      double *dwork3n, int64 nd, int64* iwork, int64 ni){
  int64 info,n;
  double *dvecwork;
  DSDPFunctionBegin;
  if (A.dsdpops->matfactor1){
    info=(A.dsdpops->matfactor1)(A.matdata); DSDPChkDataError(A,info);
  } else if (A.dsdpops->matfactor2){
    info=SDPConeVecGetSize(W,&n);
    info=SDPConeVecGetArray(W,&dvecwork);
    info=(A.dsdpops->matfactor2)(A.matdata,dworknn,nn0,dvecwork,n,dwork3n,nd,iwork,ni); DSDPChkDataError(A,info);
    info=SDPConeVecRestoreArray(W,&dvecwork);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0);   
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatDot"
/*!
\fn int64 DSDPDataMatDot(DSDPDataMat A, double x[], int64 nn, int64 n, double *v);

\brief Compute inner product of data with a dense matrix
\param A symmetric data matrix
\param x dense array matrix
\param nn length of array
\param n dimension of matrix.
\param v the inner product

\sa SDPConeSetStorageFormat()
\sa DSDPVMatGetArray()
*/
int64 DSDPDataMatDot(DSDPDataMat A, double x[], int64 nn, int64 n, double *v){
  int64 info;

  DSDPFunctionBegin;
  if (A.dsdpops->matdot){
    info=(A.dsdpops->matdot)(A.matdata,x,nn,n,v); DSDPChkDataError(A,info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatVecVec"
/*!
\fn int64 DSDPDataMatVecVec(DSDPDataMat A, SDPConeVec W, double *v);

\brief Compute w' A w
\param A symmetric data matrix
\param W vector
\param v the inner product


*/
int64 DSDPDataMatVecVec(DSDPDataMat A, SDPConeVec W, double *v){
  int64 info,n;
  double *x;

  DSDPFunctionBegin;
  if (A.dsdpops->matvecvec){
    info=SDPConeVecGetSize(W,&n); DSDPCHKERR(info);
    info=SDPConeVecGetArray(W,&x); DSDPCHKERR(info);
    info=(A.dsdpops->matvecvec)(A.matdata,x,n,v); DSDPChkDataError(A,info);
    info=SDPConeVecRestoreArray(W,&x); DSDPCHKERR(info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatMultiply"
/*
\fn int64 DSDPDataMatMultiply(DSDPDataMat A, SDPConeVec V1, SDPConeVec V2);

\brief Compute V2 = A*V1;
\param A symmetric data matrix
\param V1 in vector
\param V2 the product
Not needed.
*/
int64 DSDPDataMatMultiply(DSDPDataMat A, SDPConeVec V1, SDPConeVec V2){
  int64 info,n;
  double *vv1,*vv2;

  DSDPFunctionBegin;
  if (A.dsdpops->matmultiply){
    info=SDPConeVecGetSize(V1,&n); DSDPCHKERR(info);
    info=SDPConeVecGetArray(V1,&vv1); DSDPCHKERR(info);
    info=SDPConeVecGetArray(V2,&vv2); DSDPCHKERR(info);
    info=(A.dsdpops->matmultiply)(A.matdata,vv1,vv2,n); DSDPChkDataError(A,info);
    info=SDPConeVecRestoreArray(V1,&vv1); DSDPCHKERR(info);
    info=SDPConeVecRestoreArray(V2,&vv2); DSDPCHKERR(info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatGetRowNonzeros"
/*!
\fn int64 DSDPDataMatGetRowNonzeros(DSDPDataMat A, int64 nrow, int64 nmax, int64 *nz, int64 *nnz);

\brief Get sparsity pattern of a row of the matrix.
\param A symmetric data matrix.
\param nrow number >=0 and < n.
\param nmax dimension of the matrix.
\param nz array used to mark nonzeros.
\param nnz number of nonzeros in row.
Used to create sparse data structure for S and Delta S.
*/
int64 DSDPDataMatGetRowNonzeros(DSDPDataMat A, int64 nrow, int64 nmax, int64 *nz, int64 *nnz){
  int64 i,info;
  DSDPFunctionBegin;
  if (A.dsdpops->matrownz){
    info=(A.dsdpops->matrownz)(A.matdata,nrow,nz,nnz,nmax); DSDPChkDataError(A,info);
  } else {
    *nnz=nmax;
    for (i=0;i<nmax;i++){
      nz[i]++;
    }
  }
  DSDPFunctionReturn(0); 
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatAddRowMultipleToVector"
int64 DSDPDataMatAddRowMultipleToVector(DSDPDataMat A, int64 nrow, double ytmp, SDPConeVec R){
  int64 info,n;
  double *vv;
  DSDPFunctionBegin;
  if (A.dsdpops->mataddrowmultiple){
    info=SDPConeVecGetArray(R,&vv);DSDPCHKERR(info);
    info=SDPConeVecGetSize(R,&n);DSDPCHKERR(info);
    info=(A.dsdpops->mataddrowmultiple)(A.matdata,nrow,ytmp,vv,n); DSDPChkDataError(A,info);
    info=SDPConeVecRestoreArray(R,&vv);DSDPCHKERR(info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatAddMultiple"
/*!
\fn int64 DSDPDataMatAddMultiple(DSDPDataMat A, double ytmp, double *v, int64 nn, int64 n);

\brief Add a multiple the data matrix to the array.
\param A symmetric data matrix
\param ytmp scalar multiple
\param v dense array matrix
\param nn dimension of array
\param n dimension of matrix

\sa SDPConeSetStorageFormat()
\sa DSDPVMatGetArray()
*/
int64 DSDPDataMatAddMultiple(DSDPDataMat A, double ytmp, double *v, int64 nn, int64 n){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops->mataddallmultiple){
    info=(A.dsdpops->mataddallmultiple)(A.matdata,ytmp,v,nn,n); DSDPChkDataError(A,info);
  } else {
    DSDPNoOperationError(A);
  }
  DSDPFunctionReturn(0); 
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatView"
/*!
\fn int64 DSDPDataMatView(DSDPDataMat A);

\brief Print matrix
\param A symmetric data matrix

*/
int64 DSDPDataMatView(DSDPDataMat A){
  int64 info;
  DSDPFunctionBegin;
  if (A.dsdpops->matview){
    info=(A.dsdpops->matview)(A.matdata); DSDPChkDataError(A,info);
  } else {
    printf("No matrix view available for matrix type %s.\n",A.dsdpops->matname);
  }
  DSDPFunctionReturn(0); 
}


#undef __FUNCT__
#define __FUNCT__ "DSDPDataMatDestroy"
/*!
\fn int64 DSDPDataMatDestroy(DSDPDataMat *A);

\brief Free the data structures
\param A symmetric data matrix

*/
int64 DSDPDataMatDestroy(DSDPDataMat* A){
  int64 info;
  DSDPFunctionBegin;
  if ( (*A).dsdpops->matdestroy){
    info=((*A).dsdpops->matdestroy)((*A).matdata); DSDPChkDataError(*A,info);
  } else {
    /*   DSDPNoOperationError(*A); */
  }
  info=DSDPDataMatInitialize(A); DSDPCHKERR(info);
  /*  info=DSDPZeroMatCreate(0,A); DSDPCHKERR(info); */

  DSDPFunctionReturn(0); 
}

