#ifndef __DSDP_H
#define __DSDP_H

#include "dsdpbasictypes.h"
#include "dsdpvec.h"
#include "dsdpschurmat.h"
#include "dsdpcone.h"
#include "dsdpconverge.h"
/*!
  \file dsdp.h
  \brief Internal data structure for the DSDP solver
*/

/*!
\typedef struct LUBounds_C* YBoundCone;
\brief Cone with bounds on variables y.
*/
typedef struct LUBounds_C* YBoundCone;

/*!
\typedef typedef struct RDCone* RRCone;
\brief Cone with nonnegativity on variable r.
*/
typedef struct RDCone* RRCone;


#define MAX_DSDP_MONITORS 5
#define MAX_XMAKERS 4

typedef struct {
  DSDPVec y;
  DSDPVec dy;
  double  mu;
  double  pstep;
  DSDPVec rhs;
} XMaker;

typedef struct { /* This information is needed to compute the step Direction */
  DSDPVec y;
  double zbar;
  double mutarget;
  double logdet;
} DSDPState;

typedef struct {
  int64 (*f)(void*);
  void * ptr;
}DRoutine;

typedef struct {
  int64 (*monitor)(struct DSDP_C *, void*);
  void *monitorctx;
} DMonitor;

typedef struct {
  DSDPCone cone;
  int64 coneid;
} DCone;

/*!
\struct DSDP_C dsdp.h "src/solver/dsdp.h"
\brief Internal structures for the DSDP solver.
\sa DSDP
*/
struct DSDP_C{

  DSDPCG *sles;
  int64 slestype;

  double schurmu;
  DSDPSchurMat M;
  double Mshift;
  double maxschurshift;

  int64 ncones,maxcones;
  DCone* K;

  int64 keyid;

  int64 solvetime,cgtime,ptime,dtime,ctime;
  int64 reuseM;
  DSDPTruth goty0;
  DSDPTruth setupcalled;

  int64    m;       /*     number of constraints                        */ 
  double np;      /*     Dimension of full variable matrix            */

  int64    itnow;   /* current iterate                                  */
  int64    maxiter; /* Maximum number of iterates                       */
  double pobj;    /* current primal objective value - use duality gap */
  double ppobj;   /* current primal objetive value - evaluate P       */
  double dobj,ddobj;  /* the current dual objective value             */
  double pstep,dstep;  /* current primal and dual step lengths        */
  double dualitygap;
  double mutarget;
  double mu,muold,mu0;      /* The current  mu */
  double rho,potential,logdet,rhon;
  double pnorm;   /* the current value of ||P||                       */
  double maxtrustradius;
  double cnorm,anorm,bnorm; 
  double tracex,tracexs;
  double rgap;
  double pstepold;

  DSDPVec y;      /* dual variables                                       */
  DSDPVec y0;     /* initial dual variables                               */
  DSDPVec ytemp;  /* temporary dual variables                             */
  DSDPVec dy1;    /* search direction 1 affine direction                  */
  DSDPVec dy2;    /* search direction 2 centering direction               */
  DSDPVec dy;     /* total search direction = constant*dy1-dy2            */
  DSDPVec rhs1;   /* objective vector b to determine step direction       */
  DSDPVec rhs2;   /* barrier vector A(S^{-1}) to determine step direction */
  DSDPVec rhs;    /* right-hand side of linear system                     */
  DSDPVec rhstemp;/* temporary rhs vector                                 */
  DSDPVec b;      /* dual objective vector                                */

  /* Multiple of identity matrix added to dual        */
  double r;
  int64 rflag;
  DSDPPenalty UsePenalty;
  RRCone rcone;

  DSDPTruth usefixedrho; /*   True if fixed rho used. */
  
  XMaker xmaker[MAX_XMAKERS]; /* step direction used to create X */
  DSDPVec xmakerrhs;

  YBoundCone ybcone;
  double pinfeas; /* Infeasible in P indirectly -- neglect numerical errors */
  double perror;  /* Infeasible in P computed directly   */

  DSDPSolutionType pdfeasible;
  double dinfeastol; /* Parameter: Classify (D) as feasible */
  double pinfeastol; /* Parameter: Classify (P) as feasible */

  ConvergenceMonitor conv;
  DSDPTerminationReason reason;

  DMonitor dmonitor[MAX_DSDP_MONITORS];
  int64 nmonitors;

  DRoutine droutine[10];
  int64 ndroutines;
};

typedef struct DSDP_C PD_DSDP;

#define DSDPKEY  5432

#define DSDPValid(a) {if (!(a)||((a)->keyid!=DSDPKEY)){ DSDPSETERR(101,"DSDPERROR: Invalid DSDP object\n");}}

#include "dsdpbasictypes.h"


extern int64 DSDPCreateLUBoundsCone(DSDP, YBoundCone*);
extern int64 BoundYConeSetBounds(YBoundCone, double, double);
extern int64 BoundYConeGetBounds(YBoundCone, double*, double*);
extern int64 BoundYConeAddX(YBoundCone,double,DSDPVec,DSDPVec,DSDPVec,double*);
extern int64 BoundYConeAddS(YBoundCone,DSDPVec,DSDPVec);

#ifdef __cplusplus
extern "C" {
#endif

extern int64 DSDPComputeObjective(DSDP, DSDPVec, double *);
extern int64 DSDPComputeDY(DSDP, double, DSDPVec, double*);
extern int64 DSDPComputeNewY(DSDP, double, DSDPVec);
extern int64 DSDPComputeRHS(DSDP, double, DSDPVec);
extern int64 DSDPComputePDY1(DSDP,double,DSDPVec);
extern int64 DSDPComputePDY(DSDP, double, DSDPVec, double*);
extern int64 DSDPComputePY(DSDP,double,DSDPVec);
extern int64 DSDPComputeG(DSDP, DSDPVec,DSDPVec,DSDPVec);
extern int64 DSDPComputePNorm(DSDP, double,DSDPVec,double*);
extern int64 DSDPComputeDualityGap(DSDP, double, double*);
extern int64 DSDPSaveYForX(DSDP, double, double);
extern int64 DSDPSetY(DSDP,double,double,DSDPVec);
extern int64 DSDPComputePotential(DSDP, DSDPVec, double,double*);
extern int64 DSDPComputePotential2(DSDP, DSDPVec, double, double, double*);
extern int64 DSDPSetRR(DSDP,double);
extern int64 DSDPGetRR(DSDP,double*);
extern int64 DSDPGetMaxYElement(DSDP,double*);

extern int64 DSDPSolveDynamicRho(DSDP);
extern int64 DSDPRefineStepDirection(DSDP,DSDPVec, DSDPVec);

/* Cone operations */
extern int64 DSDPSetUpCones(DSDP);
extern int64 DSDPSetUpCones2(DSDP, DSDPVec, DSDPSchurMat);
extern int64 DSDPSchurSparsity(DSDP, int64, int64[], int64);
extern int64 DSDPComputeSS(DSDP, DSDPVec, DSDPDualFactorMatrix, DSDPTruth*);
extern int64 DSDPInvertS(DSDP);
extern int64 DSDPComputeHessian( DSDP, DSDPSchurMat,  DSDPVec, DSDPVec);
extern int64 DSDPHessianMultiplyAdd( DSDP, DSDPVec, DSDPVec);
extern int64 DSDPComputeMaxStepLength(DSDP, DSDPVec, DSDPDualFactorMatrix, double*);
extern int64 DSDPComputeLogSDeterminant(DSDP, double*);
extern int64 DSDPComputeANorm2(DSDP,DSDPVec);
extern int64 DSDPViewCones(DSDP);
extern int64 DSDPMonitorCones(DSDP,int64);
extern int64 DSDPDestroyCones(DSDP);
extern int64 DSDPPassXVectors(DSDP,double,DSDPVec,DSDPVec);
extern int64 DSDPComputeXVariables(DSDP,double,DSDPVec,DSDPVec,DSDPVec,double*);
extern int64 DSDPGetConicDimension(DSDP,double*);
extern int64 DSDPSchurSparsity(DSDP, int64, int64[], int64);

extern int64 DSDPCGSolve(DSDP,DSDPSchurMat,DSDPVec,DSDPVec,double,DSDPTruth*);
extern int64 DSDPComputeDualStepDirection(DSDP, double, DSDPVec);
extern int64 DSDPComputeDualStepDirections(DSDP);

extern int64 DSDPSetCone(DSDP,DSDPCone);

extern int64 DSDPScaleData(DSDP);
extern int64 DSDPComputeDataNorms(DSDP);

extern int64 DSDPTakeDown(DSDP);
extern int64 DSDPSetDefaultStatistics(DSDP);
extern int64 DSDPSetDefaultParameters(DSDP);
extern int64 DSDPSetDefaultMonitors(DSDP);

/* DSDP subroutines */
extern int64 DSDPInitializeVariables(DSDP);
extern int64 DSDPObjectiveGH( DSDP, DSDPSchurMat, DSDPVec);
extern int64 DSDPCheckForUnboundedObjective(DSDP, DSDPTruth*);

extern int64 DSDPCheckConvergence(DSDP,DSDPTerminationReason *);
extern int64 DSDPMonitor(DSDP);
extern int64 DSDPPrintStats(DSDP, void*); 
extern int64 DSDPPrintStatsFile(DSDP, void*);

extern int64 DSDPGetSchurMatrix(DSDP,DSDPSchurMat*);


#ifdef __cplusplus
}
#endif

extern int64 DSDPAddRCone(DSDP,RRCone*);
extern int64 RConeSetType(RRCone, DSDPPenalty);
extern int64 RConeGetRX(RRCone, double*);

extern int64 DSDPGetConvergenceMonitor(DSDP, ConvergenceMonitor**);
extern int64 DSDPDefaultConvergence(DSDP,void *);

extern int64 DSDPSetDefaultSchurMatrixStructure(DSDP);
extern int64 DSDPFixedVariablesNorm(DSDPSchurMat,DSDPVec);
extern int64 DSDPComputeFixedYX(DSDPSchurMat,DSDPVec);

extern int64 DSDPAddBCone(DSDP,DSDPVec,double);

#endif
