
#include "dsdpschurmat_impl.h"
#include "dsdpdualmat_impl.h"
#include "dsdpdsmat_impl.h"
#include "dsdpsys.h"
/*!
\file diag.c
\brief DSDPDualMat, DSDPDSMat, and the DSDPSchurMat implentations for
diagonal matrices.
*/

typedef struct {
  int64    n;
  double *val;
  int64   owndata;
} diagmat;

static int64 DiagMatCreate(int64,diagmat**);
static int64 DiagMatMult(void*,double[],double[],int64);
static int64 DiagMatGetSize(void*, int64 *);
static int64 DiagMatAddRow2(void*, int64, double, double[], int64);
static int64 DiagMatDestroy(void*);
static int64 DiagMatView(void*);
static int64 DiagMatLogDeterminant(void*, double *);

/* static int64 DiagMatScale(double *, diagmat *); */

static int64 DiagMatCreate(int64 n,diagmat**M){
  int64 info;
  diagmat *M7;

  DSDPCALLOC1(&M7,diagmat,&info);DSDPCHKERR(info);
  DSDPCALLOC2(&M7->val,double,n,&info);DSDPCHKERR(info);
  if (n>0 && M7->val==NULL) return 1;
  M7->owndata=1; M7->n=n;
  *M=M7;
  return 0;
}

static int64 DiagMatDestroy(void* AA){
  int64 info;
  diagmat* A=(diagmat*) AA;
  if (A->owndata && A->val){ DSDPFREE(&A->val,&info);DSDPCHKERR(info);}
  DSDPFREE(&A,&info);DSDPCHKERR(info);
  return 0;
}


static int64 DiagMatMult(void* AA, double x[], double y[], int64 n){

  diagmat* A=(diagmat*) AA;
  double *vv=A->val;
  int64 i;

  if (A->n != n) return 1;
  if (x==0 && n>0) return 3;
  if (y==0 && n>0) return 3;

  for (i=0; i<n; i++){
    y[i]=x[i]*vv[i];
  }  
  return 0;
}


static int64 DiagMatGetSize(void *AA, int64 *n){
  diagmat* A=(diagmat*) AA;
  *n=A->n;
  return 0;
}

static int64 DiagMatView(void* AA){
  diagmat* A=(diagmat*) AA;
  int64 i;
  for (i=0;i<A->n; i++){
    printf(" Row: %lld, Column: %lld, Value: %8.4e \n",i,i,A->val[i]);
  }
  return 0;
}

static int64 DiagMatAddRow2(void* AA, int64 nrow, double dd, double row[], int64 n){
  diagmat* A=(diagmat*) AA;
  A->val[nrow] += dd*row[nrow];
  return 0;
}


static int64 DiagMatAddElement(void*A, int64 nrow, double dd){
  diagmat* AA = (diagmat*)A;
  AA->val[nrow] += dd;
  return 0;
}

static int64 DiagMatZeros(void*A){
  diagmat* AA = (diagmat*)A;
  int64 n=AA->n;
  memset(AA->val,0,n*sizeof(double));
  return 0;
}

static int64 DiagMatSolve(void* A, double b[], double x[],int64 n){
  diagmat* AA = (diagmat*)A;
  double *v=AA->val;
  int64 i;
  for (i=0;i<n;i++){
    x[i]=b[i]/v[i];
  }
  return 0;
}

static int64 DiagMatSolve2(void* A, int64 indx[], int64 nindx, double b[], double x[],int64 n){
  diagmat* AA = (diagmat*)A;
  double *v=AA->val;
  int64 i,j;
  memset((void*)x,0,n*sizeof(double));
  for (j=0;j<nindx;j++){
    i=indx[j];
    x[i]=b[i]/v[i];
  }
  return 0;
}

static int64 DiagMatCholeskySolveBackward(void* A, double b[], double x[],int64 n){
  int64 i;
  for (i=0;i<n;i++){
    x[i]=b[i];
  }
  return 0;
}

static int64 DiagMatInvert(void *A){
  return 0;
}

static int64 DiagMatCholeskyFactor(void*A,int64 *flag){
  diagmat* AA = (diagmat*)A;
  double *v=AA->val;
  int64 i,n=AA->n;
  *flag=0;
  for (i=0;i<n;i++){
    if (v[i]<=0){ *flag=i+1; break;}
  }
  return 0;
}

static int64 DiagMatLogDeterminant(void*A, double *dd){
  diagmat* AA = (diagmat*)A;
  double d=0,*val=AA->val;
  int64 i,n=AA->n;
  for (i=0;i<n;i++){
    if (val[i]<=0) return 1;
    d+=log(val[i]);
  }
  *dd=d;
  return 0;
}

static int64 DiagMatTakeUREntriesP(void*A, double dd[], int64 nn, int64 n){
  diagmat* AA = (diagmat*)A;
  int64 i,ii;
  double *val=AA->val;
  for (i=0;i<n;i++){
    ii=(i+1)*(i+2)/2-1;
    val[i]=dd[ii];
  }
  return 0;
}
static int64 DiagMatTakeUREntriesU(void*A, double dd[], int64 nn, int64 n){
  diagmat* AA = (diagmat*)A;
  int64 i;
  double *val=AA->val;
  for (i=0;i<n;i++){
    val[i]=dd[i*n+i];
  }
  return 0;
}

static int64 DiagMatInverseAddP(void*A, double alpha, double dd[], int64 nn, int64 n){
  diagmat* AA = (diagmat*)A;
  int64 i,ii;
  double *val=AA->val;
  for (i=0;i<n;i++){
    ii=(i+1)*(i+2)/2-1;
    dd[ii]+=alpha/val[i];
  }
  return 0;
}
static int64 DiagMatInverseAddU(void*A, double alpha, double dd[], int64 nn, int64 n){
  diagmat* AA = (diagmat*)A;
  int64 i;
  double *val=AA->val;
  for (i=0;i<n;i++){
    dd[i*n+i]+=alpha/val[i];
  }
  return 0;
}

static int64 DiagMatFull(void*A, int64* dfull){
  *dfull=1;
  return 0;
}

static struct  DSDPDualMat_Ops sdmatopsp;
static struct  DSDPDualMat_Ops sdmatopsu;
static const char* diagmatname="DIAGONAL";

static int64 DiagDualOpsInitializeP(struct  DSDPDualMat_Ops* sops){
  int64 info;
  if (sops==NULL) return 0;
  info=DSDPDualMatOpsInitialize(sops); DSDPCHKERR(info);
  sops->matcholesky=DiagMatCholeskyFactor;
  sops->matsolveforward=DiagMatSolve;
  sops->matsolvebackward=DiagMatCholeskySolveBackward;
  sops->matinvert=DiagMatInvert;
  sops->matinverseadd=DiagMatInverseAddP;
  sops->matinversemultiply=DiagMatSolve2;
  sops->matseturmat=DiagMatTakeUREntriesP;
  sops->matfull=DiagMatFull;
  sops->matdestroy=DiagMatDestroy;
  sops->matgetsize=DiagMatGetSize;
  sops->matview=DiagMatView;
  sops->matlogdet=DiagMatLogDeterminant;
  sops->id=9;
  sops->matname=diagmatname;
  return 0;
}
static int64 DiagDualOpsInitializeU(struct  DSDPDualMat_Ops* sops){
  int64 info;
  if (sops==NULL) return 0;
  info=DSDPDualMatOpsInitialize(sops); DSDPCHKERR(info);
  sops->matcholesky=DiagMatCholeskyFactor;
  sops->matsolveforward=DiagMatSolve;
  sops->matsolvebackward=DiagMatCholeskySolveBackward;
  sops->matinvert=DiagMatInvert;
  sops->matinversemultiply=DiagMatSolve2;
  sops->matseturmat=DiagMatTakeUREntriesU;
  sops->matfull=DiagMatFull;
  sops->matinverseadd=DiagMatInverseAddU;
  sops->matdestroy=DiagMatDestroy;
  sops->matgetsize=DiagMatGetSize;
  sops->matview=DiagMatView;
  sops->matlogdet=DiagMatLogDeterminant;
  sops->id=9;
  sops->matname=diagmatname;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDiagDualMatCreateP"
int64 DSDPDiagDualMatCreateP(int64 n, 
			   struct  DSDPDualMat_Ops **sops1, void**smat1,
			   struct  DSDPDualMat_Ops **sops2, void**smat2){
  diagmat *AA;
  int64 info;
  DSDPFunctionBegin;

  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DiagDualOpsInitializeP(&sdmatopsp); DSDPCHKERR(info);
  *sops1=&sdmatopsp;
  *smat1=(void*)AA;

  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DiagDualOpsInitializeP(&sdmatopsp); DSDPCHKERR(info);
  *sops2=&sdmatopsp;
  *smat2=(void*)AA;
  DSDPFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DSDPDiagDualMatCreateU"
int64 DSDPDiagDualMatCreateU(int64 n, 
			   struct  DSDPDualMat_Ops **sops1, void**smat1,
			   struct  DSDPDualMat_Ops **sops2, void**smat2){
  diagmat *AA;
  int64 info;
  DSDPFunctionBegin;
  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DiagDualOpsInitializeU(&sdmatopsu); DSDPCHKERR(info);
  *sops1=&sdmatopsu;
  *smat1=(void*)AA;
  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DiagDualOpsInitializeU(&sdmatopsu); DSDPCHKERR(info);
  *sops2=&sdmatopsu;
  *smat2=(void*)AA;
  DSDPFunctionReturn(0);
}



static int64 DiagMatVecVec(void*A,double x[], int64 n, double *vv){
  diagmat* AA = (diagmat*)A;
  double *v=AA->val,vAv=0;
  int64 i;
  for (i=0;i<n;i++){
    vAv+=x[i]*x[i]*v[i];
  }
  *vv=vAv;
  return 0;
}

static int64 DDiagDSMatOpsInitP(struct  DSDPDSMat_Ops *ddiagops){
  int64 info;
  if (ddiagops==NULL) return 0;
  info=DSDPDSMatOpsInitialize(ddiagops);DSDPCHKERR(info);
  ddiagops->matseturmat=DiagMatTakeUREntriesP;
  ddiagops->matview=DiagMatView;
  ddiagops->matgetsize=DiagMatGetSize;
  ddiagops->matmult=DiagMatMult;
  ddiagops->matvecvec=DiagMatVecVec;
  ddiagops->matzeroentries=DiagMatZeros;
  ddiagops->matdestroy=DiagMatDestroy;
  ddiagops->id=9;
  ddiagops->matname=diagmatname;
  DSDPFunctionReturn(0);
}
static int64 DDiagDSMatOpsInitU(struct  DSDPDSMat_Ops *ddiagops){
  int64 info;
  if (ddiagops==NULL) return 0;
  info=DSDPDSMatOpsInitialize(ddiagops);DSDPCHKERR(info);
  ddiagops->matseturmat=DiagMatTakeUREntriesU;
  ddiagops->matview=DiagMatView;
  ddiagops->matgetsize=DiagMatGetSize;
  ddiagops->matmult=DiagMatMult;
  ddiagops->matvecvec=DiagMatVecVec;
  ddiagops->matzeroentries=DiagMatZeros;
  ddiagops->matdestroy=DiagMatDestroy;
  ddiagops->id=9;
  ddiagops->matname=diagmatname;
  DSDPFunctionReturn(0);
}

static struct  DSDPDSMat_Ops dsdiagmatopsp;
static struct  DSDPDSMat_Ops dsdiagmatopsu;

#undef __FUNCT__
#define __FUNCT__ "DSDPDiagDSMatP"
int64 DSDPCreateDiagDSMatP(int64 n,struct  DSDPDSMat_Ops* *dsmatops, void**dsmat){

  int64 info=0;
  diagmat *AA;

  DSDPFunctionBegin;
  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DDiagDSMatOpsInitP(&dsdiagmatopsp); DSDPCHKERR(info);
  *dsmatops=&dsdiagmatopsp;
  *dsmat=(void*)AA;
  DSDPFunctionReturn(0);
}
#undef __FUNCT__
#define __FUNCT__ "DSDPDiagDSMatU"
int64 DSDPCreateDiagDSMatU(int64 n,struct  DSDPDSMat_Ops* *dsmatops, void**dsmat){

  int64 info=0;
  diagmat *AA;

  DSDPFunctionBegin;
  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DDiagDSMatOpsInitU(&dsdiagmatopsu); DSDPCHKERR(info);
  *dsmatops=&dsdiagmatopsu;
  *dsmat=(void*)AA;
  DSDPFunctionReturn(0);
}

static int64 DiagRowNonzeros(void*M, int64 row, double cols[], int64 *ncols,int64 nrows){
  DSDPFunctionBegin;
  *ncols = 1;
  cols[row]=1.0;
  DSDPFunctionReturn(0);
}

static int64 DiagAddDiag(void*M, double diag[], int64 m){
  diagmat* AA = (diagmat*)M;
  double *v=AA->val;
  int64 i;
  DSDPFunctionBegin;
  for (i=0;i<m;i++){
    v[i]+=diag[i];
  }
  DSDPFunctionReturn(0);
}

static int64 DiagMultiply(void*M, double xin[], double xout[], int64 m){
  diagmat* AA = (diagmat*)M;
  double *v=AA->val;
  int64 i;
  DSDPFunctionBegin;
  for (i=0;i<m;i++){
    xout[i]+=v[i]*xin[i];
  }
  DSDPFunctionReturn(0);
}

static int64 DiagShiftDiag(void*M, double dd){
  diagmat* AA = (diagmat*)M;
  double *v=AA->val;
  int64 i,m=AA->n;
  DSDPFunctionBegin;
  for (i=0;i<m;i++){
    v[i]+=dd;
  }
  DSDPFunctionReturn(0);
}

static int64 DiagAddElement(void*M, int64 ii, double dd){
  diagmat* AA = (diagmat*)M;
  DSDPFunctionBegin;
  AA->val[ii]+=dd;
  DSDPFunctionReturn(0);
}

static int64 DiagMatOnProcessor(void*A,int64 row,int64*flag){
  *flag=1;
  return 0;
}

static int64 DiagAssemble(void*M){
  return 0;
}

static struct  DSDPSchurMat_Ops dsdpdiagschurops;

#undef __FUNCT__
#define __FUNCT__ "DSDPDiagSchurOps"
static int64 DiagSchurOps(struct  DSDPSchurMat_Ops *sops){
  int64 info;
  DSDPFunctionBegin;
  if (!sops) return 0;
  info=DSDPSchurMatOpsInitialize(sops); DSDPCHKERR(info);
  sops->matzero=DiagMatZeros;
  sops->mataddrow=DiagMatAddRow2;
  sops->mataddelement=DiagMatAddElement;
  sops->matdestroy=DiagMatDestroy;
  sops->matfactor=DiagMatCholeskyFactor;
  sops->matsolve=DiagMatSolve;
  sops->matadddiagonal=DiagAddDiag;
  sops->matshiftdiagonal=DiagShiftDiag;
  sops->mataddelement=DiagAddElement;
  sops->matscaledmultiply=DiagMultiply;
  sops->matassemble=DiagAssemble;
  sops->pmatonprocessor=DiagMatOnProcessor;
  sops->matrownonzeros=DiagRowNonzeros;
  sops->id=9;
  sops->matname=diagmatname;
  DSDPFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DSDPGetDiagSchurMat"
int64 DSDPGetDiagSchurMat(int64 n, struct  DSDPSchurMat_Ops **sops, void **data){
  int64 info=0;
  diagmat *AA;
  DSDPFunctionBegin;
  info=DiagMatCreate(n,&AA); DSDPCHKERR(info);
  info=DiagSchurOps(&dsdpdiagschurops); DSDPCHKERR(info);
  if (sops){*sops=&dsdpdiagschurops;}
  if (data){*data=(void*)AA;}
  DSDPFunctionReturn(0);
}
