#include "dsdpdatamat_impl.h"
#include "dsdpsys.h"
/*!
\file identity.c
\brief DSDPDataMat object representing a multiple of the identity matrix.
*/

typedef struct {
  int64 n;
  double dm;
} identitymat;


static int64 IdentityMatDestroy(void*);
static int64 IdentityMatView(void*);
static int64 IdentityMatVecVec(void*, double[], int64, double *);
static int64 IdentityMatDotP(void*, double[], int64, int64, double*);
static int64 IdentityMatDotF(void*, double[], int64, int64, double*);
static int64 IdentityMatGetRank(void*, int64*,int64);
static int64 IdentityMatFactor(void*);
static int64 IdentityMatGetEig(void*, int64, double*, double[], int64,int64[],int64*);
static int64 IdentityMatAddRowMultiple(void*, int64, double, double[], int64);
static int64 IdentityMatAddMultipleP(void*, double, double[], int64, int64);
static int64 IdentityMatAddMultipleF(void*, double, double[], int64, int64);
static int64 IdentityMatGetRowNnz(void*, int64, int64[], int64*, int64);

static struct  DSDPDataMat_Ops identitymatopsp;
static struct  DSDPDataMat_Ops identitymatopsf;
static int64 IdentitymatOperationsInitializeP(struct  DSDPDataMat_Ops*);
static int64 IdentitymatOperationsInitializeF(struct  DSDPDataMat_Ops*);


/*!
\fn int64 DSDPGetIdentityDataMatP(int64 n, double val, struct  DSDPDataMat_Ops** dops, void** imat)

\brief Create a sparse matrix usuable by DSDP in packed symmetric format.
\param n number of rows and columns of the matrix
\param val multiple of identity matrix.
\param dops address of a pointer to a table of function pointers
\param imat address of a pointer to an opaque data type.
*/
#undef __FUNCT__  
#define __FUNCT__ "DSDPSetIdentityP"
int64 DSDPGetIdentityDataMatP(int64 n, double val, struct  DSDPDataMat_Ops** dops, void** imat){
  int64 info;
  identitymat *AA;
  
  DSDPFunctionBegin;
  AA=(identitymat*) malloc(1*sizeof(identitymat));
  AA->dm=val;
  AA->n=n;
  info=IdentitymatOperationsInitializeP(&identitymatopsp); DSDPCHKERR(info);
  if (dops){*dops=&identitymatopsp;}
  if (imat){*imat=(void*)AA;}
  DSDPFunctionReturn(0);
}

/*!
\fn int64 DSDPGetIdentityDataMatF(int64 n, double val, struct  DSDPDataMat_Ops** dops, void** imat)

\brief Create a sparse matrix usuable by DSDP in full symmetric format.
\param n number of rows and columns of the matrix
\param val multiple of identity matrix.
\param dops address of a pointer to a table of function pointers
\param imat address of a pointer to an opaque data type.
*/
#undef __FUNCT__  
#define __FUNCT__ "DSDPSetIdentityF"
int64 DSDPGetIdentityDataMatF(int64 n, double val, struct  DSDPDataMat_Ops** dops, void** imat){
  int64 info;
  identitymat *AA;
  
  DSDPFunctionBegin;
  AA=(identitymat*) malloc(1*sizeof(identitymat));
  AA->dm=val;
  AA->n=n;
  info=IdentitymatOperationsInitializeF(&identitymatopsf); DSDPCHKERR(info);
  if (dops){*dops=&identitymatopsf;}
  if (imat){*imat=(void*)AA;}
  DSDPFunctionReturn(0);
}

static int64 IdentityMatDestroy(void* AA){
  free(AA);
  return 0;
}


static int64 IdentityMatVecVec(void* AA, double x[], int64 n, double *v){
  identitymat* A=(identitymat*)AA;
  int64 i;
  *v=0;
  for (i=0;i<n;i++){
    *v+=x[i]*x[i];
  }
  *v *= A->dm;
  return 0;
}

static int64 IdentityMatDotP(void* AA, double x[], int64 nn, int64 n, double *v){
  identitymat* A=(identitymat*)AA;
  int64 i;
  double *xx=x;
  *v=0;
  for (i=0;i<n;i++){
    *v+=*xx;
    xx+=i+2;
  }
  *v *= 2*A->dm;
  return 0;
}

static int64 IdentityMatDotF(void* AA, double x[], int64 nn, int64 n, double *v){
  identitymat* A=(identitymat*)AA;
  int64 i;
  double *xx=x;
  *v=0;
  for (i=0;i<n;i++){
    *v+=*xx;
    xx+=n+1;
  }
  *v *= 2*A->dm;
  return 0;
}

static int64 IdentityMatFNorm2(void* AA, int64 n, double *v){
  identitymat* A=(identitymat*)AA;
  *v=A->n*A->dm*A->dm;
  return 0;
}

static int64 IdentityMatView(void* AA){
  identitymat* A=(identitymat*)AA;
  printf("Multiple of Identity matrix: All Diagonal elements equal %8.8e \n",A->dm);
  return 0;
}

static int64 IdentityMatGetRank(void *AA, int64*rank, int64 n){
  identitymat* A=(identitymat*)AA;
  *rank=A->n;;
  return 0;
}

static int64 IdentityMatFactor(void*A){
  return 0;
}

static int64 IdentityMatGetEig(void*AA, int64 neig, double *eig, double v[], int64 n, int64* indx, int64 *nind){
  identitymat* A = (identitymat*)AA;

  if (neig<0 || neig>= A->n){ *eig=0; return 0;} 
  memset((void*)v,0,(A->n)*sizeof(double)); 
  v[neig]=1.0;
  indx[0]=neig;
  *nind=1;
  *eig=A->dm;  
  return 0;
}

static int64 IdentityMatGetRowNnz(void*A, int64 nrow, int64 nz[], int64 *nnzz, int64 n){
  identitymat* AA = (identitymat*)A;
  if (nrow>=0 && nrow < AA->n){
    *nnzz=1;
    nz[nrow]++;
  } else {
    *nnzz=0;    
  }
  return 0;
}

static int64 IdentityMatCountNonzeros(void*A, int64 *nnz, int64 n){
  identitymat* AA = (identitymat*)A;
  *nnz=AA->n;
  return 0;
}

static int64 IdentityMatAddRowMultiple(void*A, int64 nrow, double dd, double rrow[], int64 n){
  identitymat* AA = (identitymat*)A;
  rrow[nrow] += dd*AA->dm;
  return 0;
}

static int64 IdentityMatAddMultipleP(void*A, double dd, double vv[], int64 nn, int64 n){
  identitymat* AA = (identitymat*)A;
  double *v=vv,dm=dd*AA->dm;
  int64 i;
  for (i=0;i<n;i++){
    *v += dm;
    v+= i+2;
  }
  return 0;
}

static int64 IdentityMatAddMultipleF(void*A, double dd, double vv[], int64 nn, int64 n){
  identitymat* AA = (identitymat*)A;
  double *v=vv,dm=dd*AA->dm;
  int64 i;
  for (i=0;i<n;i++){
    *v += dm;
    v+= n+1;
  }
  return 0;
}

static const char *datamatname="MULTIPLE OF IDENTITY";

static int64 IdentitymatOperationsInitializeP(struct  DSDPDataMat_Ops* spdiagops){
  int64 info;
  if (spdiagops==NULL) return 0;
  info=DSDPDataMatOpsInitialize(spdiagops); if (info){return info;}
  spdiagops->matfactor1=IdentityMatFactor;
  spdiagops->matgetrank=IdentityMatGetRank;
  spdiagops->matgeteig=IdentityMatGetEig;
  spdiagops->matvecvec=IdentityMatVecVec;
  spdiagops->matrownz=IdentityMatGetRowNnz;
  spdiagops->matdot=IdentityMatDotP;
  spdiagops->matfnorm2=IdentityMatFNorm2;
  spdiagops->matnnz=IdentityMatCountNonzeros;
  spdiagops->mataddrowmultiple=IdentityMatAddRowMultiple;
  spdiagops->mataddallmultiple=IdentityMatAddMultipleP;
  spdiagops->matdestroy=IdentityMatDestroy;
  spdiagops->matview=IdentityMatView;
  spdiagops->id=12;
  spdiagops->matname=datamatname;
  return 0;
}

static int64 IdentitymatOperationsInitializeF(struct  DSDPDataMat_Ops* spdiagops){
  int64 info;
  if (spdiagops==NULL) return 0;
  info=DSDPDataMatOpsInitialize(spdiagops); if (info){return info;}
  spdiagops->matfactor1=IdentityMatFactor;
  spdiagops->matgetrank=IdentityMatGetRank;
  spdiagops->matgeteig=IdentityMatGetEig;
  spdiagops->matvecvec=IdentityMatVecVec;
  spdiagops->matrownz=IdentityMatGetRowNnz;
  spdiagops->matdot=IdentityMatDotF;
  spdiagops->matfnorm2=IdentityMatFNorm2;
  spdiagops->matnnz=IdentityMatCountNonzeros;
  spdiagops->mataddrowmultiple=IdentityMatAddRowMultiple;
  spdiagops->mataddallmultiple=IdentityMatAddMultipleF;
  spdiagops->matdestroy=IdentityMatDestroy;
  spdiagops->matview=IdentityMatView;
  spdiagops->id=12;
  spdiagops->matname=datamatname;
  return 0;
}
