#include "dsdpdatamat_impl.h"
#include "dsdpsys.h"
/*!
\file onemat.c
\brief DSDPDataMat object such that all elements are the same value.
*/

typedef struct {
  double cnst;
  char UPLQ;
  int64 n;
} cmat;

static int64 ConstMatDestroy(void*);
static int64 ConstMatView(void*);
static int64 ConstMatVecVec(void*, double[], int64, double *);
static int64 ConstMatDot(void*, double[],int64,int64,double *);
static int64 ConstMatGetRank(void*, int64*, int64);
static int64 ConstMatFactor(void*);
static int64 ConstMatGetEig(void*, int64, double*, double[], int64,int64[],int64*);
static int64 ConstMatRowNnz(void*, int64, int64[], int64*, int64);
static int64 ConstMatAddRowMultiple(void*, int64, double, double[], int64);
static int64 ConstMatAddMultiple(void*, double, double[], int64,int64);
static int64 ConstMatTest(void*);

static struct  DSDPDataMat_Ops constantmatops;
static int64 ConstMatOpsInitialize(struct  DSDPDataMat_Ops*);

#undef __FUNCT__  
#define __FUNCT__ "DSDPGetConstantMat"
int64 DSDPGetConstantMat(int64 n, double value, char UPLQ, struct  DSDPDataMat_Ops**mops, void**mmat){ 
  int64 info;
  cmat*AA;
  DSDPFunctionBegin;
  AA=(cmat*) malloc(1*sizeof(cmat));
  if (AA==NULL) return 1;
  AA->cnst=value;
  AA->n=n;
  AA->UPLQ=UPLQ;
  info=ConstMatOpsInitialize(&constantmatops); if(info){return 1;}
  if (mops){*mops=&constantmatops;}
  if (mmat){*mmat=(void*)AA;}
  DSDPFunctionReturn(0);
}


static int64 ConstMatDot(void* A, double x[], int64 nn, int64 n, double *v){

  cmat* AA = (cmat*)A;
  double dtmp=0.0;
  int64 i,j;

  for (i=0;i<n;i++){
    for (j=0;j<=i;j++){
      dtmp+= (x[j]);
    }
    if (AA->UPLQ=='U'){
      x=x+n;
    } else {
      x=x+i+1;
    }
  }

  *v=2*dtmp*AA->cnst;
  return 0;
}

static int64 ConstMatVecVec(void* A, double x[], int64 n, double *v){

  cmat* AA = (cmat*)A;
  double dtmp=0.0;
  int64 i;

  for (i=0; i<n; i++){
    dtmp+=x[i];
  }
  *v=dtmp*dtmp*AA->cnst;
  return 0;
}

static int64 ConstMatAddMultiple(void*A, double dd, double vv[], int64 nn, int64 n){
  cmat* AA = (cmat*)A;
  int64 i,j;
  double ddd=dd*AA->cnst;
  for (i=0;i<n;i++){
    for (j=0;j<i;j++){
      (vv[j])+=ddd;
    }
    vv[i]+=ddd;
    if (AA->UPLQ=='U'){
      vv=vv+n;
    } else {
      vv=vv+i+1;
    }
  }
  return 0;
}

static int64 ConstMatAddRowMultiple(void*A, int64 nrow, double dd, double row[], int64 n){
  cmat* AA = (cmat*)A;
  int64 i;
  double ddd=dd*AA->cnst;
  for (i=0;i<n;i++){
    row[i] += ddd;
  }
  row[nrow] -= ddd;
 return 0;
}



static int64 ConstMatFactor(void*A){
  return 0;
}

static int64 ConstMatGetRank(void *A, int64*rank, int64 n){
  *rank=1;
  return 0;
}

static int64 ConstMatGetEig(void*A, int64 neig, double *eig, double v[], int64 n, int64  indx[], int64*nind){
  cmat* AA = (cmat*)A;
  int64 i;
  if (neig!=0) return 1;
  if (neig==0){
    for (i=0;i<n;i++){ v[i]=1.0; indx[i]=i;}
    *eig=AA->cnst; *nind=n;
  } else {  /* Or return an error */
    for (i=0;i<n;i++){ v[i]=0.0; }
    *eig=0; *nind=0;
  }
  return 0;
}


static int64 ConstMatRowNnz(void*A, int64 row, int64 nz[], int64 *nnz, int64 n){
  int64 i;
  for (i=0;i<n;i++){ nz[i]++; }
  *nnz=n;
  return 0;
}

static int64 ConstMatFNorm2(void*AA, int64 n, double *fnorm2){
  cmat* A = (cmat*)AA;
  *fnorm2=A->cnst*A->cnst*n*n;
  return 0;
}

static int64 ConstMatCountNonzeros(void*A, int64 *nnz, int64 n){
  *nnz=n*n;
  *nnz=1;
  *nnz=n;
  return 0;
}


static int64 ConstMatView(void* AA){
  cmat* A = (cmat*)AA;
  printf("Every element of the matrix is the same: %10.8e\n",A->cnst);
  return 0;
}

static int64 ConstMatTest(void* AA){
  return 0;
}


static int64 ConstMatDestroy(void* A){
  if (A) free(A);
  return 0;
}

static const char *datamatname="ALL ELEMENTS THE SAME";
static int64 ConstMatOpsInitialize(struct  DSDPDataMat_Ops* cmatops){
  int64 info;
  if (cmatops==NULL) return 0;
  info=DSDPDataMatOpsInitialize(cmatops); DSDPCHKERR(info);
  cmatops->matfactor1=ConstMatFactor;
  cmatops->matgetrank=ConstMatGetRank;
  cmatops->matgeteig=ConstMatGetEig;
  cmatops->matvecvec=ConstMatVecVec;
  cmatops->matdot=ConstMatDot;
  cmatops->mataddrowmultiple=ConstMatAddRowMultiple;
  cmatops->mataddallmultiple=ConstMatAddMultiple;
  cmatops->matdestroy=ConstMatDestroy;
  cmatops->mattest=ConstMatTest;
  cmatops->matview=ConstMatView;
  cmatops->matrownz=ConstMatRowNnz;
  cmatops->matfnorm2=ConstMatFNorm2;
  cmatops->matnnz=ConstMatCountNonzeros;
  cmatops->id=14;
  cmatops->matname=datamatname;
  return 0;
}

