#include "dsdpdatamat_impl.h"
#include "dsdpsys.h"
/*! \file vech.c
\brief DSDPDataMat for sparse matrices in upper packed symmetric format 
*/

typedef struct {
  int64    neigs;
  double *eigval;
  double *an;
  int64    *cols,*nnz;
} Eigen;

/*!
struct vechmat ;

\brief A sparse symmetric data matrix is packed format.
*/
typedef struct {
  int64    nnzeros;
  const int64    *ind;
  const double *val;
  int64 ishift;
  double alpha;

  Eigen   *Eig;
  int64 factored;
  int64 owndata;
  int64    n;
} vechmat;

#define GETI(a)    (int64)(sqrt(2*(a)+0.25)-0.5)
#define GETJ(b,c)  ((b)-((c)*((c)+1))/2)

static void getij(int64 k, int64 *i,int64 *j){
  *i=GETI(k);
  *j=GETJ(k,*i);
  return;
}
/*
static void geti2(int64 k, int64 *i,int64 *j){
  int64 r,c;
  double rr=sqrt(2*k+0.25)-0.5;
  r=(int64)rr;
  c=k-(r*(r+1))/2;
  *i=r;*j=c;
  return;
}
*/
#undef __FUNCT__  
#define __FUNCT__ "CreateVechMatWData"
static int64 CreateVechMatWdata(int64 n, int64 ishift, double alpha,const int64 *ind, const double *vals, int64 nnz, vechmat **A){
  int64 info;
  vechmat* V;
  DSDPCALLOC1(&V,vechmat,&info);DSDPCHKERR(info);
  V->n=n; V->ishift=ishift, V->ind=ind; V->val=vals;V->nnzeros=nnz;
  V->alpha=alpha;
  V->owndata=0;
  V->Eig=0;
  *A=V;
  return 0;
}


static int64 VechMatAddMultiple(void* AA, double scl, double r[], int64 nn, int64 n){
  vechmat* A=(vechmat*)AA;
  int64 k,nnz=A->nnzeros;
  const int64* ind=A->ind;
  const double *val=A->val;
  double *rr=r-A->ishift;
  scl*=A->alpha;
  for (k=0; k<nnz; ++k,++ind,++val){
    *(rr+((*ind))) +=scl*(*val);
  }
  return 0;
}

static int64 VechMatDot(void* AA, double x[], int64 nn, int64 n, double *v){
  vechmat* A=(vechmat*)AA;
  int64 k,nnz=A->nnzeros; 
  const int64 *ind=A->ind;
  double vv=0, *xx=x-A->ishift;
  const double *val=A->val;
  for (k=0;k<nnz;++k,++ind,++val){
    vv+=(*val)*(*(xx+(*ind)));
  }
  *v=2*vv*A->alpha;
  return 0;
}

static int64 EigMatVecVec(Eigen*, double[], int64, double*);
static int64 VechMatGetRank(void*,int64*,int64);

static int64 VechMatVecVec(void* AA, double x[], int64 n, double *v){
  vechmat* A=(vechmat*)AA;
  int64 info,rank=n,i=0,j,k,kk;
  const int64 *ind=A->ind,ishift=A->ishift;
  double vv=0,dd;
  const double *val=A->val,nnz=A->nnzeros;
  
  if (A->factored==3){
    info=VechMatGetRank(AA,&rank,n);
    if (nnz>3 && rank<nnz){
      info=EigMatVecVec(A->Eig,x,n,&vv);
      *v=vv*A->alpha;
      return 0;
    }
  }
  
  for (k=0; k<nnz; ++k,++ind,++val){
    kk=*ind-ishift;
    i=GETI(kk);
    j=GETJ(kk,i);
    dd=x[i]*x[j]*(*val);
    vv+=2*dd;
    if (i==j){ vv-=dd; }
  }
  *v=vv*A->alpha;

  return 0;
}


static int64 VechMatGetRowNnz(void* AA, int64 trow, int64 nz[], int64 *nnzz,int64 nn){
  vechmat* A=(vechmat*)AA;
  int64 i=0,j,k,ishift=A->ishift,nnz=A->nnzeros;
  const int64 *ind=A->ind;
  *nnzz=0;
  for (k=0; k<nnz; ++k,++ind){
    getij((*ind)-ishift,&i,&j);
    if (i==trow){
      nz[j]++;(*nnzz)++;
    } else if (j==trow){
      nz[i]++;(*nnzz)++;
    }
  }
  return 0;
}

static int64 VechMatFNorm2(void* AA, int64 n, double *fnorm2){
  vechmat* A=(vechmat*)AA;
  int64 i=0,j,k,ishift=A->ishift,nnz=A->nnzeros;
  const int64 *ind=A->ind;
  double fn2=0;
  const double *val=A->val;
  for (k=0; k<nnz; ++k,++ind){
    getij((*ind)-ishift,&i,&j);
    if (i==j){
      fn2+= val[k]*val[k];
    } else {
      fn2+= 2*val[k]*val[k];
    }
  }
  *fnorm2=fn2*A->alpha*A->alpha;
  return 0;
}

static int64 VechMatAddRowMultiple(void* AA, int64 trow, double scl, double r[], int64 m){
  vechmat* A=(vechmat*)AA;
  int64 i=0,j,k,ishift=A->ishift,nnz=A->nnzeros;
  const int64 *ind=A->ind;
  const double *val=A->val;
  scl*=A->alpha;
  for (k=0; k<nnz; ++k,++ind){
    getij((*ind)-ishift,&i,&j);
    if (i==trow){
      /*     if (i==j){ r[j]+=scl*val[k];} */
      r[j]+=scl*val[k];
    } else if (j==trow){
      r[i]+=scl*val[k];
    }
  }
  return 0;
}

static int64 VechMatCountNonzeros(void* AA, int64*nnz, int64 n){
  vechmat* A=(vechmat*)AA;
  *nnz=A->nnzeros;
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "VechMatDestroy"
static int64 VechMatDestroy(void* AA){
  vechmat* A=(vechmat*)AA;
  int64 info;
  if (A->owndata){
    /* Never happens 
    if (A->ind){ DSDPFREE(&A->ind,&info);DSDPCHKERR(info);}
    if (A->val){ DSDPFREE(&A->val,&info);DSDPCHKERR(info);}
    */
    return 1;
  }
  if (A->Eig){
    DSDPFREE(&A->Eig->eigval,&info);DSDPCHKERR(info);
    DSDPFREE(&A->Eig->an,&info);DSDPCHKERR(info);
    if (A->Eig->cols){DSDPFREE(&A->Eig->cols,&info);DSDPCHKERR(info);}
    if (A->Eig->nnz){DSDPFREE(&A->Eig->nnz,&info);DSDPCHKERR(info);}
    DSDPFREE(&A->Eig,&info);DSDPCHKERR(info);
  }
  DSDPFREE(&A,&info);DSDPCHKERR(info);
  return 0;
}



#undef __FUNCT__  
#define __FUNCT__ "DSDPCreateVechMatEigs"
static int64 CreateEigenLocker(Eigen **EE,int64 iptr[], int64 neigs, int64 n){
  int64 i,k,info;
  Eigen *E;

  for (k=0,i=0;i<neigs;i++) k+=iptr[i];
  if (k>n*neigs/4){

    DSDPCALLOC1(&E,Eigen,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->eigval,double,neigs,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->an,double,n*neigs,&info);DSDPCHKERR(info);
    E->neigs=neigs;
    E->cols=0;
    E->nnz=0;

  } else {

    DSDPCALLOC1(&E,Eigen,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->eigval,double,neigs,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->nnz,int64,neigs,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->an,double,k,&info);DSDPCHKERR(info);
    DSDPCALLOC2(&E->cols,int64,k,&info);DSDPCHKERR(info);
    E->neigs=neigs;

    if (neigs>0) E->nnz[0]=iptr[0];
    for (i=1;i<neigs;i++){E->nnz[i]=E->nnz[i-1]+iptr[i];}
  }
  *EE=E;
  return 0;
}


static int64 EigMatSetEig(Eigen* A,int64 row, double eigv, int64 idxn[], double v[], int64 nsub,int64 n){
  int64 j,k,*cols=A->cols;
  double *an=A->an;

  A->eigval[row]=eigv;
  if (cols){
    k=0; if (row>0){ k=A->nnz[row-1];}
    cols+=k; an+=k;
    for (k=0,j=0; j<nsub; j++){
      if (v[j]==0.0) continue;
      cols[k]=idxn[j]; an[k]=v[j]; k++;
    }
  } else {
    an+=n*row;
    for (j=0; j<nsub; j++){
      if (v[j]==0.0) continue;
      an[idxn[j]]=v[j];
    }
  }
  return 0;
}


static int64 EigMatGetEig(Eigen* A,int64 row, double *eigenvalue, double eigenvector[], int64 n, int64 spind[], int64 *nind){
  int64 i,*cols=A->cols,bb,ee;
  double* an=A->an;
  *eigenvalue=A->eigval[row];
  *nind=0;
  if (cols){
    memset((void*)eigenvector,0,n*sizeof(double));
    if (row==0){ bb=0;} else {bb=A->nnz[row-1];} ee=A->nnz[row];
    for (i=bb;i<ee;i++){
      eigenvector[cols[i]]=an[i];
      spind[i-bb]=cols[i]; (*nind)++;
    }
  } else {
    memcpy((void*)eigenvector,(void*)(an+n*row),n*sizeof(double));
    for (i=0;i<n;i++)spind[i]=i;
    *nind=n;
  }
  return 0;
}

static int64 EigMatVecVec(Eigen* A, double v[], int64 n, double *vv){
  int64 i,rank,*cols=A->cols,neigs=A->neigs,*nnz=A->nnz,bb,ee;
  double* an=A->an,*eigval=A->eigval,dd,ddd=0;

  if (cols){
    for (rank=0;rank<neigs;rank++){
      if (rank==0){ bb=0;} else {bb=nnz[rank-1];} ee=nnz[rank];
      for (dd=0,i=bb;i<ee;i++){
	dd+=an[i]*v[cols[i]];
      }
      ddd+=dd*dd*eigval[rank];
    }
  } else {
    for (rank=0;rank<neigs;rank++){
      for (dd=0,i=0;i<n;i++){
	dd+=an[i]*v[i];
      }
      an+=n;
      ddd+=dd*dd*eigval[rank];
    }
  }
  *vv=ddd;
  return 0;
}


static int64 VechMatComputeEigs(vechmat*,double[],int64,double[],int64,double[],int64,int64[],int64,double[],int64,double[],int64);

static int64 VechMatFactor(void*AA, double dmatp[], int64 nn0, double dwork[], int64 n, double ddwork[], int64 n1, int64 iptr[], int64 n2){
  vechmat*  A=(vechmat*)AA;
  int64 i,j,k,ishift=A->ishift,isdiag,nonzeros=A->nnzeros,info;
  int64 nn1=0,nn2=0;
  double *ss1=0,*ss2=0;
  const int64 *ind=A->ind;

  if (A->factored) return 0;
  memset((void*)iptr,0,3*n*sizeof(int64));  
  /* Find number of nonzeros in each row */
  for (isdiag=1,k=0; k<nonzeros; k++){
    getij(ind[k]-ishift,&i,&j);
    iptr[i]++;
    if (i!=j) {isdiag=0;iptr[j]++;}
  }
  
  if (isdiag){ A->factored=1; return 0;}
  /* Find most nonzeros per row */
  for (j=0,i=0; i<n; i++){ if (iptr[i]>j) j=iptr[i]; }
  if (j<2){ A->factored=2; return 0; }
  info=VechMatComputeEigs(A,dmatp,nn0,dwork,n,ddwork,n1,iptr,n2,ss1,nn1,ss2,nn2);DSDPCHKERR(info);
  A->factored=3;
  return 0;
}

static int64 VechMatGetRank(void *AA,int64 *rank,int64 n){
  vechmat*  A=(vechmat*)AA;
  switch (A->factored){
  case 1:
    *rank=A->nnzeros;
    break;
  case 2:
    *rank=2*A->nnzeros;
    break;
  case 3:
    *rank=A->Eig->neigs;
    break;
  default:
    DSDPSETERR(1,"Vech Matrix not factored yet\n");
  }
  return 0;
}

static int64 VechMatGetEig(void* AA, int64 rank, double *eigenvalue, double vv[], int64 n, int64 indx[], int64 *nind){
  vechmat*  A=(vechmat*)AA;
  const double *val=A->val,tt=sqrt(0.5);
  int64 info,i,j,k,ishift=A->ishift;
  const int64 *ind=A->ind;

  *nind=0;
  switch (A->factored){
  case 1:
    memset(vv,0,n*sizeof(double));
    getij(ind[rank]-ishift,&i,&j);
    vv[i]=1.0;
    *eigenvalue=val[rank]*A->alpha;
    *nind=1;
    indx[0]=i;
    break;
  case 2:
    memset(vv,0,n*sizeof(double));
    k=rank/2;
    getij(ind[k]-ishift,&i,&j);
    if (i==j){
      if (k*2==rank){ 
	vv[i]=1.0; *eigenvalue=val[k]*A->alpha;
	*nind=1;
	indx[0]=i;
      } else {
	*eigenvalue=0;
      }
    } else {
      if (k*2==rank){
	vv[i]=tt;  vv[j]=tt; *eigenvalue=val[k]*A->alpha;
	*nind=2;
	indx[0]=i; indx[1]=j;
      } else {
	vv[i]=-tt; vv[j]=tt; *eigenvalue=-val[k]*A->alpha;
	*nind=2;
	indx[0]=i; indx[1]=j;
      }
    }
    break;
  case 3:
    info=EigMatGetEig(A->Eig,rank,eigenvalue,vv,n,indx,nind);DSDPCHKERR(info);
    *eigenvalue=*eigenvalue*A->alpha;
    break;
  default:
    DSDPSETERR(1,"Vech Matrix not factored yet\n");
  }

  return 0;  
}

static int64 VechMatView(void* AA){
  vechmat* A=(vechmat*)AA;
  int64 info,i=0,j,k,rank=0,ishift=A->ishift,nnz=A->nnzeros;
  const int64 *ind=A->ind;
  const double *val=A->val;
  for (k=0; k<nnz; k++){
    getij(ind[k]-ishift,&i,&j);
    printf("Row: %lld, Column: %lld, Value: %10.8f \n",i,j,A->alpha*val[k]);
  }
  if (A->factored>0){
    info=VechMatGetRank(AA,&rank,A->n);DSDPCHKERR(info);
    printf("Detected Rank: %lld\n",rank);
  }
  return 0;
}


static struct  DSDPDataMat_Ops vechmatops;
static const char *datamatname="STANDARD VECH MATRIX";

static int64 VechMatOpsInitialize(struct  DSDPDataMat_Ops *sops){
  int64 info;
  if (sops==NULL) return 0;
  info=DSDPDataMatOpsInitialize(sops); DSDPCHKERR(info);
  sops->matvecvec=VechMatVecVec;
  sops->matdot=VechMatDot;
  sops->matfnorm2=VechMatFNorm2;
  sops->mataddrowmultiple=VechMatAddRowMultiple;
  sops->mataddallmultiple=VechMatAddMultiple;
  sops->matview=VechMatView;
  sops->matdestroy=VechMatDestroy;
  sops->matfactor2=VechMatFactor;
  sops->matgetrank=VechMatGetRank;
  sops->matgeteig=VechMatGetEig;
  sops->matrownz=VechMatGetRowNnz;
  sops->matnnz=VechMatCountNonzeros;
  sops->id=3;
  sops->matname=datamatname;
  return 0;
}

/*!
\fn  int64 DSDPGetVechMat(int64 n,int64 ishift,double alpha, const int64 ind[], const double val[],int64 nnz, struct  DSDPDataMat_Ops**sops, void**smat)
\brief Given data in packed symmetric format, create a sparse matrix usuable by DSDP.
\param n number of rows and columns of the matrix
\param ishift the index of the first element in the matrix (usually 0)
\param alpha the multiple of these matrix.
\param ind array of indices for matrix.
\param val array of matrix values.
\param nnz size of arrays.
\param sops address of a pointer to a table of function pointers
\param smat address of a pointer to an opaque data type.
*/
#undef __FUNCT__  
#define __FUNCT__ "DSDPGetVechMat"
int64 DSDPGetVechMat(int64 n,int64 ishift,double alpha, const int64 ind[], const double val[],int64 nnz, struct  DSDPDataMat_Ops**sops, void**smat){ 
  int64 info,i,j,k,itmp,nn=n*(n+1)/2;
  double dtmp;
  vechmat* AA;
  DSDPFunctionBegin;
  for (k=0;k<nnz;++k){ 
    itmp=ind[k]-ishift; 
    if (itmp>=nn){
      getij(itmp,&i,&j);
      /*
      DSDPSETERR(2,"Illegal index value: Element %d in array has row %d (>0) or column %d (>0) is greater than %d. \n",k+1,i+1,j+1,n);
      */
      DSDPSETERR3(2,"Illegal index value: Element %d in array has index %d greater than or equal to %d. \n",k,itmp,nn);
    } else if (itmp<0){
      DSDPSETERR1(2,"Illegal index value: %d.  Must be >= 0\n",itmp);
    }
  }
  for (k=0;k<nnz;++k) dtmp=val[k];
  info=CreateVechMatWdata(n,ishift,alpha,ind,val,nnz,&AA); DSDPCHKERR(info);
  AA->factored=0;
  AA->Eig=0;
  info=VechMatOpsInitialize(&vechmatops); DSDPCHKERR(info);
  if (sops){*sops=&vechmatops;}
  if (smat){*smat=(void*)AA;}
  DSDPFunctionReturn(0);
}


#undef __FUNCT__  
#define __FUNCT__ "VechMatComputeEigs"
static int64 VechMatComputeEigs(vechmat* AA,double DD[], int64 nn0, double W[], int64 n, double WORK[], int64 n1, int64 iiptr[], int64 n2, double ss1[],int64 nn1, double ss2[], int64 nn2){

  int64 i,j,k,nsub,neigs,info,*iptr,*perm,*invp;
  int64 *i2darray=(int64*)DD;
  int64 ownarray1=0,ownarray2=0,ownarray3=0;
  int64 ishift=AA->ishift,nonzeros=AA->nnzeros;
  const int64 *ind=AA->ind;
  const double *val=AA->val;
  double *dmatarray=ss1,*dworkarray=ss2,maxeig,eps=1.0e-12,eps2=1.0e-12;
  
  iptr=iiptr;  perm=iptr+n;  invp=perm+n;
  /* These operations were done before calling this routine * /
  / * Integer arrays corresponding to rows with nonzeros and inverse map * /
  memset((void*)iiptr,0,3*n*sizeof(int64));

  / * Find number of nonzeros in each row * /
  for (i=0,k=0; k<nonzeros; k++){
    getij(ind[k],i,n,&i,&j);
    iptr[i]++; iptr[j]++;
  } 
  */
  /* Number of rows with a nonzero.  Order the rows with nonzeros. */
  for (nsub=0,i=0; i<n; i++){
    if (iptr[i]>0){ invp[nsub]=i; perm[i]=nsub; nsub++;}
  }
  
  /* create a dense array in which to put numbers */
  if (nsub*nsub>nn1){
    DSDPCALLOC2(&dmatarray,double,(nsub*nsub),&info); DSDPCHKERR(info);
    ownarray1=1;
  }
  memset((void*)dmatarray,0,nsub*nsub*sizeof(double));
  if (nsub*nsub>nn2){
    DSDPCALLOC2(&dworkarray,double,(nsub*nsub),&info); DSDPCHKERR(info);
    ownarray2=1;
  }

  if (nsub*nsub*sizeof(int64)>nn0*sizeof(double)){
    DSDPCALLOC2(&i2darray,int64,(nsub*nsub),&info); DSDPCHKERR(info);
    ownarray3=1;
  }
  
  for (i=0,k=0; k<nonzeros; k++){
    getij(ind[k]-ishift,&i,&j);
    dmatarray[perm[i]*nsub+perm[j]] += val[k];
    if (i!=j){
      dmatarray[perm[j]*nsub+perm[i]] += val[k];
    }
  }
  /* Call LAPACK to compute the eigenvalues */
  info=DSDPGetEigs(dmatarray,nsub,dworkarray,nsub*nsub,i2darray,nsub*nsub,
		   W,n,WORK,n1,iiptr+3*n,n2-3*n);
  if (info){
    memset((void*)dmatarray,0,nsub*nsub*sizeof(double));
    for (i=0,k=0; k<nonzeros; k++){
      getij(ind[k]-ishift,&i,&j);
      dmatarray[perm[i]*nsub+perm[j]] += val[k];
      if (i!=j){dmatarray[perm[j]*nsub+perm[i]] += val[k];}
    }
    info=DSDPGetEigs2(dmatarray,nsub,dworkarray,nsub*nsub,i2darray,nsub*nsub,
		      W,nsub,WORK,n1,iiptr+3*n,n2-3*n); DSDPCHKERR(info);
  } 
  for (maxeig=0,i=0;i<nsub;i++){
    if (fabs(W[i])>maxeig){ maxeig=fabs(W[i]); }
  }
  memset((void*)iptr,0,nsub*sizeof(int64));
  /* Compute sparsity pattern for  eigenvalue and eigenvector structures */
  /* Count the nonzero eigenvalues */
  for (neigs=0,k=0; k<nsub; k++){
    if (fabs(W[k]) /* /maxeig */ > eps){
      for (j=0;j<nsub;j++){
	if (fabs(dmatarray[nsub*k+j]) >= eps2){iptr[neigs]++;
	} else { dmatarray[nsub*k+j]=0.0;}
      }
      neigs++;
      /*
    } else if (fabs(W[k])>1.0e-100){
      printf("SKIPPING EIGENVALUE: %4.4e, max is : %4.4e\n",W[k],maxeig);
      */
    } 
  }
  
  info=CreateEigenLocker(&AA->Eig,iptr,neigs,n);DSDPCHKERR(info);
  DSDPLogInfo(0,49," Data Mat has %d eigenvalues: \n",neigs);
  /* Copy into structure */
  for (neigs=0,i=0; i<nsub; i++){
    if (fabs(W[i]) > eps){
      info=EigMatSetEig(AA->Eig,neigs,W[i],invp,dmatarray+nsub*i,nsub,n);DSDPCHKERR(info);
      neigs++;
    }
  }

  if (ownarray1){ DSDPFREE(&dmatarray,&info);DSDPCHKERR(info);}
  if (ownarray2){ DSDPFREE(&dworkarray,&info);DSDPCHKERR(info);}
  if (ownarray3){ DSDPFREE(&i2darray,&info);DSDPCHKERR(info);}
  return 0;
}
 
